# from .ReadHelper import read_int_8
# from io import BytesIO
#
# import compoundfiles
# import zlib
#
#
# def swizzle(b):
#     b ^= 0xD2
#     b <<= 1
#     b |= b >> 8
#     return b & 0xFF
#
#
# def parse_art_file(file):
#     with open(file, 'rb') as f:
#         contents = compoundfiles.CompoundFileReader(f).open('Contents')
#         contents.seek(4)  # file size
#         return zlib.decompress(bytes([swizzle(b) for b in contents.read()]))
#
#
# def read(f, out, settings=None):
#     contents = compoundfiles.CompoundFileReader(f).open('Contents')
#     contents.seek(4)  # file size
#     art = BytesIO(zlib.decompress(bytes([swizzle(b) for b in contents.read()])))
#     with open('file.bin', 'bw') as f:
#         while True:
#             b = read_int_8(art)
#             if b is None or b == -1:
#                 break
#             f.write(bytearray(chr(b), 'utf-8'))
#             print('%02X' % b)



from .EmbThread import EmbThread
from .ReadHelper import read_int_16le, read_int_8, read_int_32be, signed8, signed16


# 7F 08 00 00 is color change.
# 7F 01 xx yy is unstitched.
# 7F 7F 02 14 is end.

def read(f, out, settings=None):
    f.seek(0x27, 1)
    num_of_colors = read_int_16le(f)
    f.seek(0x100, 0)
    while True:
        b1 = read_int_8(f)
        if b1 == 0x7D or b1 == 0x7E:  # not seen a 7E big jump code
            x = read_int_16le(f)
            y = read_int_16le(f)
            out.move(signed16(x), -signed16(y))
            continue
        b2 = read_int_8(f)
        if b1 != 0x7F:
            out.stitch(signed8(b1), -signed8(b2))
            continue
        b3 = read_int_8(f)
        b4 = read_int_8(f)
        if b2 == 0x01:
            out.move(signed8(b3), -signed8(b4))
            continue
        elif b2 == 0x03:
            out.trim()
            x = signed8(b3)
            y = -signed8(b4)
            if x != 0 or y != 0:
                out.move(x, y)
            continue
        elif b2 == 0x08 or 0x0A <= b2 <= 0x17:
            out.color_change()
            continue
        elif b2 == 0x7F:
            break  # End
        elif b2 == 0x18:
            break  # End
    out.end()
    f.seek(2, 1)
    for i in range(0, num_of_colors):
        thread = EmbThread()
        thread.color = read_int_32be(f)
        if thread.color is None:
            break
        else:
            out.add_thread(thread)



from .EmbThread import EmbThread
from .ReadHelper import read_int_8, read_int_24be, read_int_24le, read_int_16le, signed24

PC_SIZE_CONVERSION_RATIO = 5.0 / 3.0


def read_pc_file(f, out, settings=None):
    version = read_int_8(f)
    hoop_size = read_int_8(f)
    # 0 for PCD,
    # 1 for PCQ (MAXI),
    # 2 for PCS small hoop(80x80),
    # 3 for PCS with large hoop.
    color_count = read_int_16le(f)
    for i in range(0, color_count):
        thread = EmbThread()
        thread.color = read_int_24be(f)
        out.add_thread(thread)
        f.seek(1, 1)

    stitch_count = read_int_16le(f)
    while True:
        c0 = read_int_8(f)
        x = read_int_24le(f)
        c1 = read_int_8(f)
        y = read_int_24le(f)
        ctrl = read_int_8(f)
        if ctrl is None:
            break
        x = signed24(x)
        y = -signed24(y)
        x *= PC_SIZE_CONVERSION_RATIO
        y *= PC_SIZE_CONVERSION_RATIO
        if ctrl == 0x00:
            out.stitch_abs(x, y)
            continue
        if ctrl & 0x01:
            out.color_change()
            continue
        if ctrl & 0x04:
            out.move_abs(x, y)
            continue
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    read_pc_file(f, out)



from .ReadHelper import read_int_8, read_int_32le, read_int_24le, signed24

MAX_SIZE_CONVERSION_RATIO = 1.235


def read(f, out, settings=None):
    f.seek(0xD5, 0)
    stitch_count = read_int_32le(f)
    for i in range(0, stitch_count):
        x = read_int_24le(f)
        c0 = read_int_8(f)
        y = read_int_24le(f)
        c1 = read_int_8(f)
        x = signed24(x)
        y = signed24(y)
        x *= MAX_SIZE_CONVERSION_RATIO
        y *= MAX_SIZE_CONVERSION_RATIO
        if c1 is None:
            break
        out.stitch_abs(x, y)
    out.end()



from __future__ import print_function

import unittest

from pattern_for_tests import *


class TestColorFormats(unittest.TestCase):

    def test_write_read_col(self):
        file1 = "color.col"
        for m in range(0,50):
            pattern = EmbPattern()
            for i in range(4, 20):
                pattern.add_thread(EmbThread("random"))
            write(pattern, file1)
            w_pattern = read(file1)
            self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
            for q in range(0,len(pattern.threadlist)):
                self.assertEqual(pattern.threadlist[q].color, w_pattern.threadlist[q].color)
        self.addCleanup(os.remove, file1)

    def test_write_read_edr(self):
        file1 = "color.edr"
        for m in range(0, 50):
            pattern = EmbPattern()
            for i in range(4, 20):
                pattern.add_thread(EmbThread("random"))
            write(pattern, file1)
            w_pattern = read(file1)
            self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
            for q in range(0, len(pattern.threadlist)):
                self.assertEqual(pattern.threadlist[q].color, w_pattern.threadlist[q].color)
        self.addCleanup(os.remove, file1)

    def test_write_read_inf(self):
        file1 = "color.inf"
        for m in range(0, 50):
            pattern = EmbPattern()
            for i in range(4, 20):
                pattern.add_thread(EmbThread("random"))
            write(pattern, file1)
            w_pattern = read(file1)
            self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
            for q in range(0, len(pattern.threadlist)):
                self.assertEqual(pattern.threadlist[q].color, w_pattern.threadlist[q].color)
        self.addCleanup(os.remove, file1)

    def test_write_edr_dst(self):
        pattern = get_shift_pattern()
        file1 = "color.edr"
        file2 = "color.dst"
        write(pattern, file1)
        write(pattern, file2)
        w_pattern = read(file1)
        w_pattern = read(file2, pattern=w_pattern)
        self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
        for q in range(0, len(pattern.threadlist)):
            self.assertEqual(pattern.threadlist[q].color, w_pattern.threadlist[q].color)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_write_dst_edr(self):
        pattern = get_shift_pattern()
        file1 = "color.dst"
        file2 = "color.edr"
        write(pattern, file1)
        write(pattern, file2)
        w_pattern = read(file1)
        w_pattern = read(file2, pattern=w_pattern)
        self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
        for q in range(0, len(pattern.threadlist)):
            self.assertEqual(pattern.threadlist[q].color, w_pattern.threadlist[q].color)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_write_dst_col(self):
        pattern = get_shift_pattern()
        file1 = "color.dst"
        file2 = "color.col"
        write(pattern, file1)
        write(pattern, file2)
        w_pattern = read(file1)
        w_pattern = read(file2, pattern=w_pattern)
        self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
        for q in range(0, len(pattern.threadlist)):
            self.assertEqual(pattern.threadlist[q].color, w_pattern.threadlist[q].color)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_write_dst_inf(self):
        pattern = get_shift_pattern()
        file1 = "color.dst"
        file2 = "color.inf"
        write(pattern, file1)
        write(pattern, file2)
        w_pattern = read(file1)
        w_pattern = read(file2, pattern=w_pattern)
        self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
        for q in range(0, len(pattern.threadlist)):
            self.assertEqual(pattern.threadlist[q].color, w_pattern.threadlist[q].color)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)


from .EmbConstant import *


def encode_thread_change(command, thread=None, needle=None, order=None):
    if thread is None:
        thread = 0
    else:
        thread &= 0xFF
        thread += 1
    if needle is None:
        needle = 0
    else:
        needle &= 0xFF
        needle += 1
    if order is None:
        order = 0
    else:
        order &= 0xFF
        order += 1
    command &= COMMAND_MASK
    return command | (order << 24) | (needle << 16) | (thread << 8)


def decode_embroidery_command(command):
    flag = command & COMMAND_MASK
    thread = command & THREAD_MASK
    thread >>= 8
    thread -= 1
    if thread == -1:
        thread = None
    needle = command & NEEDLE_MASK
    needle >>= 16
    needle -= 1
    if needle == -1:
        needle = None
    order = command & ORDER_MASK
    order >>= 24
    order -= 1
    if order == -1:
        order = None
    return flag, thread, needle, order


def get_command_dictionary():
    return {
        "NO_COMMAND": NO_COMMAND,
        "STITCH": STITCH,
        "JUMP": JUMP,
        "TRIM": TRIM,
        "STOP": STOP,
        "END": END,
        "SLOW": SLOW,
        "FAST": FAST,
        "COLOR_CHANGE": COLOR_CHANGE,
        "NEEDLE_SET": NEEDLE_SET,
        "SET_CHANGE_SEQUENCE": SET_CHANGE_SEQUENCE,
        "SEQUIN_MODE": SEQUIN_MODE,
        "SEQUIN_EJECT": SEQUIN_EJECT,
        "SEW_TO": SEW_TO,
        "NEEDLE_AT": NEEDLE_AT,
        "STITCH_BREAK": STITCH_BREAK,
        "SEQUENCE_BREAK": SEQUENCE_BREAK,
        "COLOR_BREAK": COLOR_BREAK,
        "TIE_ON": TIE_ON,
        "TIE_OFF": TIE_OFF,
        "FRAME_EJECT": FRAME_EJECT,
        "MATRIX_TRANSLATE": MATRIX_TRANSLATE,
        "MATRIX_SCALE_ORIGIN": MATRIX_SCALE_ORIGIN,
        "MATRIX_ROTATE_ORIGIN": MATRIX_ROTATE_ORIGIN,
        "MATRIX_SCALE": MATRIX_SCALE,
        "MATRIX_ROTATE": MATRIX_ROTATE,
        "MATRIX_RESET": MATRIX_RESET,
        "CONTINGENCY_TIE_ON_THREE_SMALL": CONTINGENCY_TIE_ON_THREE_SMALL,
        "CONTINGENCY_TIE_OFF_THREE_SMALL": CONTINGENCY_TIE_OFF_THREE_SMALL,
        "CONTINGENCY_TIE_ON_NONE": CONTINGENCY_TIE_ON_NONE,
        "CONTINGENCY_TIE_OFF_NONE": CONTINGENCY_TIE_OFF_NONE,
        "OPTION_MAX_STITCH_LENGTH": OPTION_MAX_STITCH_LENGTH,
        "OPTION_MAX_JUMP_LENGTH": OPTION_MAX_JUMP_LENGTH,
        "OPTION_IMPLICIT_TRIM": OPTION_IMPLICIT_TRIM,
        "OPTION_EXPLICIT_TRIM": OPTION_EXPLICIT_TRIM,
        "CONTINGENCY_LONG_STITCH_NONE": CONTINGENCY_LONG_STITCH_NONE,
        "CONTINGENCY_LONG_STITCH_JUMP_NEEDLE": CONTINGENCY_LONG_STITCH_JUMP_NEEDLE,
        "CONTINGENCY_LONG_STITCH_SEW_TO": CONTINGENCY_LONG_STITCH_SEW_TO,
    }


def get_common_name_dictionary():
    return {
        NO_COMMAND: "NO_COMMAND",
        STITCH: "STITCH",
        JUMP: "JUMP",
        TRIM: "TRIM",
        STOP: "STOP",
        END: "END",
        SLOW: "SLOW",
        FAST: "FAST",
        COLOR_CHANGE: "COLOR_CHANGE",
        NEEDLE_SET: "NEEDLE_SET",
        SET_CHANGE_SEQUENCE: "SET_CHANGE_SEQUENCE",
        SEQUIN_MODE: "SEQUIN_MODE",
        SEQUIN_EJECT: "SEQUIN_EJECT",
        SEW_TO: "SEW_TO",
        NEEDLE_AT: "NEEDLE_AT",
        STITCH_BREAK: "STITCH_BREAK",
        SEQUENCE_BREAK: "SEQUENCE_BREAK",
        COLOR_BREAK: "COLOR_BREAK",
        TIE_ON: "TIE_ON",
        TIE_OFF: "TIE_OFF",
        FRAME_EJECT: "FRAME_EJECT",
        MATRIX_TRANSLATE: "MATRIX_TRANSLATE",
        MATRIX_SCALE: "MATRIX_SCALE",
        MATRIX_ROTATE: "MATRIX_ROTATE",
        MATRIX_SCALE_ORIGIN: "MATRIX_SCALE_ORIGIN",
        MATRIX_ROTATE_ORIGIN: "MATRIX_ROTATE_ORIGIN",
        MATRIX_RESET: "MATRIX_RESET",
        CONTINGENCY_TIE_ON_THREE_SMALL: "CONTINGENCY_TIE_ON_THREE_SMALL",
        CONTINGENCY_TIE_OFF_THREE_SMALL: "CONTINGENCY_TIE_OFF_THREE_SMALL",
        CONTINGENCY_TIE_ON_NONE: "CONTINGENCY_TIE_ON_NONE",
        CONTINGENCY_TIE_OFF_NONE: "CONTINGENCY_TIE_OFF_NONE",
        OPTION_MAX_STITCH_LENGTH: "OPTION_MAX_STITCH_LENGTH",
        OPTION_MAX_JUMP_LENGTH: "OPTION_MAX_JUMP_LENGTH",
        OPTION_IMPLICIT_TRIM: "OPTION_IMPLICIT_TRIM",
        OPTION_EXPLICIT_TRIM: "OPTION_EXPLICIT_TRIM",
        CONTINGENCY_LONG_STITCH_NONE: "CONTINGENCY_LONG_STITCH_NONE",
        CONTINGENCY_LONG_STITCH_JUMP_NEEDLE: "CONTINGENCY_LONG_STITCH_JUMP_NEEDLE",
        CONTINGENCY_LONG_STITCH_SEW_TO: "CONTINGENCY_LONG_STITCH_SEW_TO"
    }



from .EmbThread import EmbThread
from .ReadHelper import read_int_8


def read(f, out, settings=None):
    while True:
        red = read_int_8(f)
        green = read_int_8(f)
        blue = read_int_8(f)
        if blue is None:
            return
        f.seek(1, 1)
        thread = EmbThread()
        thread.set_color(red, green, blue)
        out.add_thread(thread)



from .EmbConstant import *
from .WriteHelper import write_int_8, write_int_24be, \
    write_int_32be, write_int_16be, write_string_utf8, write_string

SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_JUMP
FULL_JUMP = False
# Since jump is stitch, full jump will result in pointless double stitch.
# Vp3 can encode signed 16 bit deltas.
MAX_JUMP_DISTANCE = 3200
# coded beyond 255 tho, they count as jumps.
MAX_STITCH_DISTANCE = 255


def vp3_write_string_8(stream, string):
    write_int_16be(stream, len(string))
    write_string_utf8(stream, string)
    # vp3_write_length_and_bytes(stream, bytestring)


def vp3_write_string_16(stream, string):
    write_int_16be(stream, len(string) * 2)
    write_string(stream, string, "UTF-16BE")
    # vp3_write_length_and_bytes(stream, bytestring)


def vp3_write_length_and_bytes(stream, bytestring):
    write_int_16be(stream, len(bytestring))
    stream.write(bytestring)


def vp3_patch_byte_offset(stream, offset):
    current_pos = stream.tell()
    stream.seek(offset, 0)  # Absolute position seek.
    position = current_pos - offset - 4  # 4 bytes int32
    write_int_32be(stream, position)
    stream.seek(current_pos, 0)  # Absolute position seek.


def get_as_colorblocks(pattern):
    thread_index = 0
    last_pos = 0
    end = len(pattern.stitches)
    for pos, stitch in enumerate(pattern.stitches):
        command = stitch[2] & COMMAND_MASK
        if command != COLOR_CHANGE:
            continue
        thread = pattern.get_thread_or_filler(thread_index)
        thread_index += 1
        yield (pattern.stitches[last_pos:pos], thread)
        last_pos = pos
    thread = pattern.get_thread_or_filler(thread_index)
    thread_index += 1
    yield (pattern.stitches[last_pos:end], thread)


def write(pattern, f, settings=None):
    pattern.fix_color_count()

    write_string_utf8(f, "%vsm%")
    write_int_8(f, 0)
    vp3_write_string_16(f, "Produced by     Software Ltd")
    write_file(pattern, f)


def write_file(pattern, f):
    f.write(b'\x00\x02\x00')
    placeholder_distance_end_of_file_block_020 = f.tell()
    write_int_32be(f, 0)  # placeholder
    # This refers to the end of the final block, not entire bytes.

    vp3_write_string_16(f, "")
    # This is global notes and settings string.
    # "Setting:" followed by settings text.

    count_stitches = len(pattern.stitches)
    colorblocks = [i for i in get_as_colorblocks(pattern)]

    count_colorblocks_total = len(colorblocks)

    extends = pattern.bounds()
    write_int_32be(f, int(extends[2] * 100))  # right
    write_int_32be(f, int(extends[1] * -100))  # -top
    write_int_32be(f, int(extends[0] * 100))  # left
    write_int_32be(f, int(extends[3] * -100))  # -bottom

    # EmbroiderModder Comment:
    # "this would be some(unknown) function of thread length"
    # Wilcom: 0C 54 == 3156
    # Note, this is the total stitch count, sans end command.
    ends = pattern.count_stitch_commands(END)
    count_just_stitches = count_stitches - ends

    write_int_32be(f, count_just_stitches)
    write_int_8(f, 0)
    write_int_8(f, count_colorblocks_total)
    write_int_8(f, 12)  # 0xC
    write_int_8(f, 0)

    count_designs = 1
    write_int_8(f, count_designs)  # Number of designs.
    for i in range(0, count_designs):
        write_design_block(f, extends, colorblocks)
    vp3_patch_byte_offset(f, placeholder_distance_end_of_file_block_020)


def write_design_block(f, extends, colorblocks):
    f.write(b'\x00\x03\x00')
    placeholder_distance_end_of_design_block_030 = f.tell()
    write_int_32be(f, 0)

    count_colorblocks_total = len(colorblocks)

    width = extends[2] - extends[0]
    height = extends[3] - extends[1]
    half_width = width / 2
    half_height = height / 2
    center_x = extends[2] - half_width
    center_y = extends[3] - half_height

    write_int_32be(f, int(center_x) * 100)  # initial x
    write_int_32be(f, int(center_y) * -100)  # initial y
    write_int_8(f, 0)
    write_int_8(f, 0)
    write_int_8(f, 0)

    # extends 2
    write_int_32be(f, int(half_width) * -100)
    write_int_32be(f, int(half_width) * 100)
    write_int_32be(f, int(half_height) * -100)
    write_int_32be(f, int(half_height) * 100)

    write_int_32be(f, int(width) * 100)
    write_int_32be(f, int(height) * 100)
    vp3_write_string_16(f, "")  # This is notes and settings string.

    f.write(b'\x64\x64')  # write_int_16be(f, 25700)
    # maybe b'dd', maybe 100, 100
    write_int_32be(f, 4096)  # b'\x00\x00\x10\x00'
    write_int_32be(f, 0)  # b'\x00\x00\x00\x00'
    write_int_32be(f, 0)  # b'\x00\x00\x10\x00'
    write_int_32be(f, 4096)  # b'\x00\x00\x10\x00'

    f.write(b'xxPP\x01\x00')

    vp3_write_string_16(f, "Produced by     Software Ltd")

    write_int_16be(f, count_colorblocks_total)

    first = True
    for colorblock in colorblocks:
        stitches = colorblock[0]
        thread = colorblock[1]
        write_vp3_colorblock(f, first, center_x, center_y, stitches, thread)
        first = False
    vp3_patch_byte_offset(f, placeholder_distance_end_of_design_block_030)


def write_vp3_colorblock(f, first, center_x, center_y, stitches, thread):
    f.write(b'\x00\x05\x00')
    placeholder_distance_end_of_color_block_050 = f.tell()
    write_int_32be(f, 0)

    if len(stitches) > 0:
        first_pos_x = stitches[0][0]
        first_pos_y = stitches[0][1]
        if first:
            first_pos_x = 0
            first_pos_y = 0
        last_pos_x = stitches[-1][0]
        last_pos_y = stitches[-1][1]
    else:
        first_pos_x = 0
        first_pos_y = 0
        last_pos_x = 0
        last_pos_y = 0
    start_position_from_center_x = first_pos_x - center_x
    start_position_from_center_y = -(first_pos_y - center_y)
    write_int_32be(f, int(start_position_from_center_x) * 100)
    write_int_32be(f, int(start_position_from_center_y) * 100)

    vp3_write_thread(f, thread)

    block_shift_x = last_pos_x - first_pos_x
    block_shift_y = -(last_pos_y - first_pos_y)

    write_int_32be(f, int(block_shift_x) * 100)
    write_int_32be(f, int(block_shift_y) * 100)

    write_stitches_block(f, stitches, first_pos_x, first_pos_y)

    write_int_8(f, 0)
    vp3_patch_byte_offset(f, placeholder_distance_end_of_color_block_050)


def vp3_write_thread(f, thread):
    f.write(b'\x01\x00')  # Single color, no transition.
    write_int_24be(f, thread.color)
    f.write(b'\x00\x00\x00\x05\x28')  # no parts, no length, Rayon 40-weight
    if thread.catalog_number is not None:
        vp3_write_string_8(f, thread.catalog_number)
    else:
        vp3_write_string_8(f, "")
    if thread.description is not None:
        vp3_write_string_8(f, thread.description)
    else:
        vp3_write_string_8(f, thread.hex_color())
    if thread.brand is not None:
        vp3_write_string_8(f, thread.brand)
    else:
        vp3_write_string_8(f, "")


def write_stitches_block(f, stitches, first_pos_x, first_pos_y):
    # The 0, x, 0 bytes come before placeholders
    f.write(b'\x00\x01\x00')
    placeholder_distance_to_end_of_stitches_block_010 = f.tell()
    write_int_32be(f, 0)  # placeholder

    f.write(b'\x0A\xF6\x00')
    last_x = first_pos_x
    last_y = first_pos_y

    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        flags = stitch[2] & COMMAND_MASK
        alt = stitch[2] & FLAGS_MASK
        if flags == END:
            # This is a trim command. The machine does not autotrim.
            # Consequently writers tend to add this explicit trim command.
            f.write(b'\x80\x03')
            break
        elif flags == COLOR_CHANGE:
            # Colorchange commands divided the pattern into colorblocks.
            continue
        elif flags == TRIM:
            f.write(b'\x80\x03')
            continue
        elif flags == SEQUIN_MODE:
            continue
        elif flags == SEQUIN_EJECT:
            continue
        elif flags == STOP:
            continue
        elif flags == JUMP:
            # VP3 has no jump commands. These are skipped.
            # It moves to the relevant location without needing to block the needlebar.
            continue
        dx = int(x - last_x)
        dy = int(y - last_y)
        last_x += dx
        last_y += dy
        if flags == STITCH:
            if -127 <= dx <= 127 and -127 <= dy <= 127 and alt == 0:
                write_int_8(f, dx)
                write_int_8(f, dy)
            else:
                f.write(b'\x80\x01')
                write_int_16be(f, dx)
                write_int_16be(f, dy)
                f.write(b'\x80\x02')
    vp3_patch_byte_offset(f, placeholder_distance_to_end_of_stitches_block_010)



from .EmbThreadPec import get_thread_set
from .ReadHelper import read_string_8, read_int_8, read_int_24le

JUMP_CODE = 0x10
TRIM_CODE = 0x20
FLAG_LONG = 0x80


def read(f, out, settings=None):
    pec_string = read_string_8(f, 8)
    # pec_string must equal #PEC0001
    read_pec(f, out)
    out.interpolate_duplicate_color_as_stop()


def read_pec(f, out, pes_chart=None):
    f.seek(3, 1)  # LA:
    label = read_string_8(f, 16)  # Label
    if label is not None:
        out.metadata("Name", label.strip())
    f.seek(0xF, 1)  # Dunno, spaces then 0xFF 0x00
    pec_graphic_byte_stride = read_int_8(f)
    pec_graphic_icon_height = read_int_8(f)
    f.seek(0xC, 1)
    color_changes = read_int_8(f)
    count_colors = color_changes + 1  # PEC uses cc - 1, 0xFF means 0.
    color_bytes = bytearray(f.read(count_colors))
    threads = []
    map_pec_colors(color_bytes, out, pes_chart, threads)
    f.seek(0x1D0 - color_changes, 1)
    stitch_block_end = read_int_24le(f) - 5 + f.tell()
    # The end of this value is already 5 into the stitchblock.

    # 3 bytes, '\x31\xff\xf0', 6 2-byte shorts. 15 total.
    f.seek(0x0F, 1)
    read_pec_stitches(f, out)
    f.seek(stitch_block_end, 0)

    byte_size = pec_graphic_byte_stride * pec_graphic_icon_height

    read_pec_graphics(f,
                      out,
                      byte_size,
                      pec_graphic_byte_stride,
                      count_colors + 1,
                      threads
                      )


def read_pec_graphics(f, out, size, stride, count, values):
    v = values[:]
    v.insert(0, None)
    for i in range(0, count):
        graphic = bytearray(f.read(size))
        if f is not None:
            name = "pec_graphic_" + str(i)
            out.metadata(name, (graphic, stride, v[i]))


def process_pec_colors(colorbytes, out, values):
    thread_set = get_thread_set()
    max_value = len(thread_set)
    for byte in colorbytes:
        thread_value = thread_set[byte % max_value]
        out.add_thread(thread_value)
        values.append(thread_value)


def process_pec_table(colorbytes, out, chart, values):
    # This is how PEC actually allocates pre-defined threads to blocks.
    thread_set = get_thread_set()
    max_value = len(thread_set)
    thread_map = {}
    for i in range(0, len(colorbytes)):
        color_index = int(colorbytes[i] % max_value)
        thread_value = thread_map.get(color_index, None)
        if thread_value is None:
            if len(chart) > 0:
                thread_value = chart.pop(0)
            else:
                thread_value = thread_set[color_index]
            thread_map[color_index] = thread_value
        out.add_thread(thread_value)
        values.append(thread_value)


def map_pec_colors(colorbytes, out, chart, values):
    if chart is None or len(chart) == 0:
        # Reading pec colors.
        process_pec_colors(colorbytes, out, values)

    elif len(chart) >= len(colorbytes):
        # Reading threads in 1 : 1 mode.
        for thread in chart:
            out.add_thread(thread)
            values.append(thread)
    else:
        # Reading tabled mode threads.
        process_pec_table(colorbytes, out, chart, values)


def signed12(b):
    b &= 0xFFF
    if b > 0x7FF:
        return - 0x1000 + b
    else:
        return b


def signed7(b):
    if b > 63:
        return - 128 + b
    else:
        return b


def read_pec_stitches(f, out):
    while True:
        val1 = read_int_8(f)
        val2 = read_int_8(f)
        if (val1 == 0xFF and val2 == 0x00) or val2 is None:
            break
        if val1 == 0xFE and val2 == 0xB0:
            f.seek(1, 1)
            out.color_change(0, 0)
            continue
        jump = False
        trim = False
        if val1 & FLAG_LONG != 0:
            if val1 & TRIM_CODE != 0:
                trim = True
            if val1 & JUMP_CODE != 0:
                jump = True
            code = (val1 << 8) | val2
            x = signed12(code)
            val2 = read_int_8(f)
            if val2 is None:
                break
        else:
            x = signed7(val1)

        if val2 & FLAG_LONG != 0:
            if val2 & TRIM_CODE != 0:
                trim = True
            if val2 & JUMP_CODE != 0:
                jump = True
            val3 = read_int_8(f)
            if val3 is None:
                break
            code = val2 << 8 | val3
            y = signed12(code)
        else:
            y = signed7(val2)
        if jump:
            out.move(x, y)
        elif trim:
            out.trim()
            out.move(x, y)
        else:
            out.stitch(x, y)
    out.end()



from .EmbConstant import *
from .WriteHelper import write_string_utf8, write_int_8, write_int_16le

MAX_STITCH_DISTANCE = 70
MAX_PERMITTED_STITCHES = 100


def write(pattern, f, settings=None):
    max_x = -200000000
    min_x = +200000000
    max_y = -200000000
    min_y = +200000000
    point_count = 0
    for stitch in pattern.stitches:
        data = stitch[2]
        x = stitch[0]
        y = -stitch[1]
        if data == STITCH or data == JUMP:
            point_count += 1
            if x > max_x:
                max_x = x
            if x < min_x:
                min_x = x
            if y > max_y:
                max_y = y
            if y < min_y:
                min_y = y
        if point_count >= MAX_PERMITTED_STITCHES:
            break
    center_y = (min_y + max_y) / 2.0
    normal_max_y = max_y - center_y
    if normal_max_y > 35.0:  # 14 * 2.5 = 35.0
        scale_y = 14.0 / normal_max_y  # 1 / (normal_max_y / 14.0)
    else:
        scale_y = 1.0 / 2.5  # pure unit conversion.
    scale_x = 1.0 / 2.5

    write_string_utf8(f, "#PMV0001")
    header = "...................................."
    write_string_utf8(f, header[0:36])
    f.write(b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00')

    write_int_16le(f, point_count)
    write_int_16le(f, point_count * 2)
    point_index = -1
    max_x = -200000000
    min_x = +200000000
    max_y = -200000000
    min_y = +200000000
    xx = 0
    for stitch in pattern.stitches:
        point_index += 1
        if point_index >= point_count:
            break
        data = stitch[2] & COMMAND_MASK
        x = stitch[0]
        y = -stitch[1]
        x *= scale_x
        y -= center_y
        y *= scale_y
        y = int(round(y))
        x = int(round(x - xx))
        xx += x
        if data == STITCH or data == JUMP:
            if xx > max_x:
                max_x = xx
            if xx < min_x:
                min_x = xx
            if y > max_y:
                max_y = y
            if y < min_y:
                min_y = y

            if x < 0:
                x += 64
            if y < 0:
                y += 32
            write_int_8(f, x)
            write_int_8(f, y)
            continue
    write_int_16le(f, 0)
    write_int_16le(f, 256)
    f.write(b'\x00\x00\x00\x00\x05\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x02\x00')
    write_int_16le(f, 256)
    write_int_8(f, 0)
    write_int_8(f, 0)
    length_range = max_x - min_x
    write_length_lookup_table(f, length_range)
    width_range = max_y - min_y
    write_width_lookup_table(f, width_range)
    write_int_16le(f, 0x12)
    f.write(b'\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00')


def write_length_lookup_table(f, length_range):
    # I've not solved this for how they are actually made, writing a something that should work.
    write_values = [(0, 0), (10, 71), (20, 143), (40, 214), (60, 286), (80, 357),
                    (100, 429), (120, 500), (140, 571), (160, 714), (180, 786), (200, 857),
                    (250, 1000), (300, 1286), (350, 1429), (400, 1571), (450, 1786), (500, 2000)]

    steps = len(write_values)
    write_int_8(f, steps-1)  # (500, 2000)
    write_int_8(f, steps)
    for value in write_values:
        length_at_step = value[0]
        other_at_step = value[1]
        write_int_16le(f, length_at_step)
        write_int_16le(f, other_at_step)


def write_width_lookup_table(f, width_range):
    if width_range == 0:
        write_int_8(f, 0)
        write_int_8(f, 1)
        write_int_16le(f, 8192)
        write_int_16le(f, 1000)
        return
    steps = 15
    second_max = 28000.0 / float(width_range)
    second_step = second_max / float(steps - 1)
    write_int_8(f, steps-1)
    write_int_8(f, steps)
    for i in range(0, steps):
        width_at_step = 50 * i
        other_at_step = second_step * i

        write_int_16le(f, width_at_step)
        write_int_16le(f, int(round(other_at_step)))



from .EmbConstant import *


def read_u01_stitches(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        ctrl = byte[0]
        dy = -byte[1]
        dx = byte[2]
        if (ctrl & 0x20) != 0:
            dx = -dx
        if (ctrl & 0x40) != 0:
            dy = -dy
        command = ctrl & 0b11111
        if command == 0x0:
            # Stitch
            out.stitch(dx, dy)
            continue
        if command == 0x01:
            # Jump
            out.move(dx, dy)
            continue
        if command == 0x02:
            # Fast
            out.add_stitch_relative(FAST)
            if dx != 0 or dy != 0:
                out.stitch(dx, dy)
            continue
        if command == 0x03:
            # Fast, Jump
            out.add_stitch_relative(FAST)
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x04:
            # Slow
            out.add_stitch_relative(SLOW)
            if dx != 0 or dy != 0:
                out.stitch(dx, dy)
            continue
        if command == 0x05:
            # Slow, Jump
            out.add_stitch_relative(SLOW)
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x06:
            # T1 Top Thread Trimming, TTrim.
            out.trim()
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x07:
            # T2 Bobbin Threading
            out.trim()
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x08:  # ww, stop file had proper A8 rather than E8 and displacement
            # C00 Stop
            out.stop()
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if 0x09 <= command <= 0x17:
            # C01 - C14
            needle = command - 0x08
            out.needle_change(needle)
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x18:
            break
        if ctrl == 0x2B:
            break  # Rare postfix data from machine. Do not read this.
        break  # Uncaught Command
    out.end()


def read(f, out, settings=None):
    f.seek(0x80, 1)
    f.seek(0x80, 1)
    read_u01_stitches(f, out)



from __future__ import print_function

import unittest
from pyembroidery import *


class TestMatrix(unittest.TestCase):

    def test_matrix(self):
        matrix = EmbMatrix()
        matrix.post_rotate(90, 100, 100)
        p = matrix.point_in_matrix_space(50, 50)
        self.assertAlmostEqual(p[0], 150)
        self.assertAlmostEqual(p[1], 50)

    def test_matrix_2(self):
        matrix = EmbMatrix()
        matrix.reset()
        matrix.post_scale(2, 2, 50, 50)
        p = matrix.point_in_matrix_space(50, 50)
        self.assertAlmostEqual(p[0], 50)
        self.assertAlmostEqual(p[1], 50)

        p = matrix.point_in_matrix_space(25, 25)
        self.assertAlmostEqual(p[0], 0)
        self.assertAlmostEqual(p[1], 0)
        matrix.post_rotate(45, 50, 50)

        p = matrix.point_in_matrix_space(25, 25)
        self.assertAlmostEqual(p[0], 50)

    def test_matrix_3(self):
        matrix = EmbMatrix()
        matrix.reset()
        matrix.post_scale(0.5, 0.5)
        p = matrix.point_in_matrix_space(100, 100)
        self.assertAlmostEqual(p[0], 50)
        self.assertAlmostEqual(p[1], 50)
        matrix.reset()
        matrix.post_scale(2, 2, 100, 100)
        p = matrix.point_in_matrix_space(50, 50)
        self.assertAlmostEqual(p[0], 0)
        self.assertAlmostEqual(p[1], 0)

    def test_matrix_rotate(self):
        pattern = EmbPattern()
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "red")
        pattern.add_command(MATRIX_ROTATE, 45)
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "blue")
        pattern.add_command(MATRIX_ROTATE, 45)
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "aqua")
        pattern = pattern.get_normalized_pattern()
        print("rotate:", pattern.stitches)
        self.assertAlmostEqual(pattern.stitches[4][0], pattern.stitches[6][0])
        self.assertAlmostEqual(pattern.stitches[4][1], pattern.stitches[6][1])
        self.assertAlmostEqual(pattern.stitches[10][0], pattern.stitches[12][0])
        self.assertAlmostEqual(pattern.stitches[10][1], pattern.stitches[12][1])
        self.assertAlmostEqual(pattern.stitches[4][0], pattern.stitches[12][0])
        self.assertAlmostEqual(pattern.stitches[4][1], pattern.stitches[12][1])
        file1 = "file.svg"
        write_svg(pattern, file1)
        self.addCleanup(os.remove, file1)

    def test_matrix_translate(self):
        pattern = EmbPattern()
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "red")
        pattern.add_command(MATRIX_TRANSLATE, 20, 40)
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "blue")
        pattern.add_command(MATRIX_TRANSLATE, -20, -40)
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "aqua")
        pattern = pattern.get_normalized_pattern()
        print("translate:", pattern.stitches)
        self.assertIsNotNone(pattern.stitches)
        self.assertEqual(pattern.count_stitch_commands(MATRIX_TRANSLATE), 0)

        self.assertAlmostEqual(pattern.stitches[4][0], pattern.stitches[12][0])
        self.assertAlmostEqual(pattern.stitches[4][1], pattern.stitches[12][1])
        file1 = "file2.svg"
        write_svg(pattern, file1)
        self.addCleanup(os.remove, file1)

    def test_matrix_translate_rotate(self):
        pattern = EmbPattern()
        pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "red")
        pattern.add_command(MATRIX_TRANSLATE, 20, 40)
        pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "blue")
        pattern.add_command(MATRIX_ROTATE, -90)
        pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "aqua")
        pattern = pattern.get_normalized_pattern()
        print("transrot:", pattern.stitches)
        self.assertIsNotNone(pattern.stitches)
        self.assertEqual(pattern.count_stitch_commands(MATRIX_TRANSLATE), 0)
        self.assertEqual(pattern.count_stitch_commands(MATRIX_ROTATE), 0)
        self.assertAlmostEqual(pattern.stitches[12][0], pattern.stitches[6][0])
        self.assertAlmostEqual(pattern.stitches[12][1], pattern.stitches[6][1])
        file1 = "file3.svg"
        write_svg(pattern, file1)
        self.addCleanup(os.remove, file1)

    def test_matrix_translate_scale(self):
        pattern = EmbPattern()
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "red")
        pattern.add_command(MATRIX_TRANSLATE, 20, 40)
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "blue")
        pattern.add_command(MATRIX_SCALE, 2, 2)
        pattern.add_block([(10, 10), (10, 110), (110, 110), (110, 10), (10, 10)], "aqua")
        pattern = pattern.get_normalized_pattern()
        print("transcale:", pattern.stitches)
        self.assertIsNotNone(pattern.stitches)
        self.assertEqual(pattern.count_stitch_commands(MATRIX_TRANSLATE), 0)
        self.assertEqual(pattern.count_stitch_commands(MATRIX_SCALE), 0)
        self.assertAlmostEqual(pattern.stitches[12][0], pattern.stitches[6][0])
        self.assertAlmostEqual(pattern.stitches[12][1], pattern.stitches[6][1])
        file1 = "file4.svg"
        write_svg(pattern, file1)
        self.addCleanup(os.remove, file1)



from .EmbConstant import *
from .WriteHelper import write_string_utf8

SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_UTILIZE
FULL_JUMP = False
ROUND = True
MAX_JUMP_DISTANCE = 121
MAX_STITCH_DISTANCE = 121

PPMM = 10
DSTHEADERSIZE = 512


def bit(b):
    return 1 << b


def encode_record(x, y, flags):
    y = -y  # flips the coordinate y space.
    b0 = 0
    b1 = 0
    b2 = 0
    if flags == JUMP or flags == SEQUIN_EJECT:
        b2 += bit(7)  # jumpstitch 10xxxx11
    if flags == STITCH or flags == JUMP or flags == SEQUIN_EJECT:
        b2 += bit(0)
        b2 += bit(1)
        if x > 40:
            b2 += bit(2)
            x -= 81
        if x < -40:
            b2 += bit(3)
            x += 81
        if x > 13:
            b1 += bit(2)
            x -= 27
        if x < -13:
            b1 += bit(3)
            x += 27
        if x > 4:
            b0 += bit(2)
            x -= 9
        if x < -4:
            b0 += bit(3)
            x += 9
        if x > 1:
            b1 += bit(0)
            x -= 3
        if x < -1:
            b1 += bit(1)
            x += 3
        if x > 0:
            b0 += bit(0)
            x -= 1
        if x < 0:
            b0 += bit(1)
            x += 1
        if x != 0:
            raise ValueError("The dx value given to the writer exceeds maximum allowed.")
        if y > 40:
            b2 += bit(5)
            y -= 81
        if y < -40:
            b2 += bit(4)
            y += 81
        if y > 13:
            b1 += bit(5)
            y -= 27
        if y < -13:
            b1 += bit(4)
            y += 27
        if y > 4:
            b0 += bit(5)
            y -= 9
        if y < -4:
            b0 += bit(4)
            y += 9
        if y > 1:
            b1 += bit(7)
            y -= 3
        if y < -1:
            b1 += bit(6)
            y += 3
        if y > 0:
            b0 += bit(7)
            y -= 1
        if y < 0:
            b0 += bit(6)
            y += 1
        if y != 0:
            raise ValueError("The dy value given to the writer exceeds maximum allowed.")
    elif flags == COLOR_CHANGE:
        b2 = 0b11000011
    elif flags == STOP:
        b2 = 0b11000011
    elif flags == END:
        b2 = 0b11110011
    elif flags == SEQUIN_MODE:
        b2 = 0b01000011
    return bytes(bytearray([b0, b1, b2]))


def write(pattern, f, settings=None):
    extended_header = False
    trim_at = 3
    if settings is not None:
        extended_header = settings.get("extended header", extended_header)  # deprecated, use version="extended"
        version = settings.get("version", "default")
        if version == "extended":
            extended_header = True
        trim_at = settings.get("trim_at", trim_at)
    bounds = pattern.bounds()

    name = pattern.get_metadata("name", "Untitled")

    write_string_utf8(f, "LA:%-16s\r" % name)
    write_string_utf8(f, "ST:%7d\r" % pattern.count_stitches())
    write_string_utf8(f, "CO:%3d\r" % pattern.count_color_changes())

    write_string_utf8(f, "+X:%5d\r" % abs(bounds[2]))
    write_string_utf8(f, "-X:%5d\r" % abs(bounds[0]))
    write_string_utf8(f, "+Y:%5d\r" % abs(bounds[3]))
    write_string_utf8(f, "-Y:%5d\r" % abs(bounds[1]))
    ax = 0
    ay = 0
    if len(pattern.stitches) > 0:
        last = len(pattern.stitches) - 1
        ax = int(pattern.stitches[last][0])
        ay = -int(pattern.stitches[last][1])
    if ax >= 0:
        write_string_utf8(f, "AX:+%5d\r" % ax)
    else:
        write_string_utf8(f, "AX:-%5d\r" % abs(ax))
    if ay >= 0:
        write_string_utf8(f, "AY:+%5d\r" % ay)
    else:
        write_string_utf8(f, "AY:-%5d\r" % abs(ay))
    write_string_utf8(f, "MX:+%5d\r" % 0)
    write_string_utf8(f, "MY:+%5d\r" % 0)
    write_string_utf8(f, "PD:%6s\r" % "******")
    if extended_header:
        author = pattern.get_metadata("author")
        if author is not None:
            write_string_utf8(f, "AU:%s\r" % author)
        meta_copyright = pattern.get_metadata("copyright")
        if meta_copyright is not None:
            write_string_utf8(f, "CP:%s\r" % meta_copyright)
        if len(pattern.threadlist) > 0:
            for thread in pattern.threadlist:
                write_string_utf8(f, "TC:%s,%s,%s\r" %
                                  (thread.hex_color(),
                                   thread.description,
                                   thread.catalog_number))
    f.write(b'\x1a')
    for i in range(f.tell(), DSTHEADERSIZE):
        f.write(b'\x20')  # space

    stitches = pattern.stitches
    xx = 0
    yy = 0
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2] & COMMAND_MASK
        dx = int(round(x - xx))
        dy = int(round(y - yy))

        xx += dx
        yy += dy
        if data == TRIM:
            delta = -4
            f.write(encode_record(-delta/2, -delta/2, JUMP))
            for p in range(1,trim_at-1):
                f.write(encode_record(delta, delta, JUMP))
                delta = -delta
            f.write(encode_record(delta/2, delta/2, JUMP))
        else:
            f.write(encode_record(dx, dy, data))



from .EmbThread import EmbThread
from .ReadHelper import read_int_32be, read_int_16be


def read(f, out, settings=None):
    u0 = read_int_32be(f)
    u1 = read_int_32be(f)
    u2 = read_int_32be(f)
    number_of_colors = read_int_32be(f)
    for j in range(0, number_of_colors):
        length = read_int_16be(f) - 2  # 2 bytes of the length.
        byte_data = bytearray(f.read(length))
        if len(byte_data) != length:
            break
        red = byte_data[2]
        green = byte_data[3]
        blue = byte_data[4]
        thread = EmbThread()
        thread.set_color(red, green, blue)
        byte_data = byte_data[7:]
        for j in range(0, len(byte_data)):
            b = byte_data[j]
            if b == 0:
                thread.description = byte_data[:j].decode('utf8')
                byte_data = byte_data[j+1:]
                break
        for j in range(0, len(byte_data)):
            b = byte_data[j]
            if b == 0:
                thread.chart = byte_data[:j].decode('utf8')
                break
        out.add_thread(thread)



from .EmbThreadJef import get_thread_set
from .ReadHelper import read_int_32le, signed8


def read_jef_stitches(f, out, settings=None):
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] != 0x80:
            x = signed8(b[0])
            y = -signed8(b[1])
            out.stitch(x, y)
            continue
        ctrl = b[1]
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        x = signed8(b[0])
        y = -signed8(b[1])
        if ctrl == 0x02:
            out.move(x, y)
            continue
        if ctrl == 0x01:
            out.color_change(0, 0)
            continue
        if ctrl == 0x10:
            break
        break  # Uncaught Control
    out.end(0, 0)

    clipping = True
    trims = False
    count_max = None
    trim_distance = 3.0
    if settings is not None:
        count_max = settings.get('trim_at', count_max)
        trims = settings.get("trims", trims)
        trim_distance = settings.get("trim_distance", trim_distance)
        clipping = settings.get('clipping', clipping)
    if trims and count_max is None:
        count_max = 3
    if trim_distance is not None:
        trim_distance *= 10 # Pixels per mm. Native units are 1/10 mm.
    out.interpolate_trims(count_max, trim_distance, clipping)


def read(f, out, settings=None):
    jef_threads = get_thread_set()
    stitch_offset = read_int_32le(f)
    f.seek(20, 1)
    count_colors = read_int_32le(f)
    f.seek(88, 1)

    for i in range(0, count_colors):
        index = abs(read_int_32le(f))
        out.add_thread(jef_threads[index % len(jef_threads)])

    f.seek(stitch_offset, 0)
    read_jef_stitches(f, out, settings)



from .ReadHelper import read_int_8, read_int_16le


def find_extends(stitches):
    min_x = float('inf')
    min_y = float('inf')
    max_x = -float('inf')
    max_y = -float('inf')

    for stitch in stitches:
        if stitch[0] > max_x:
            max_x = stitch[0]
        if stitch[0] < min_x:
            min_x = stitch[0]
        if stitch[1] > max_y:
            max_y = stitch[1]
        if stitch[1] < min_y:
            min_y = stitch[1]
    return min_x, min_y, max_x, max_y


def read_pmv_stitches(f, out, settings=None):
    """PMV files are stitch files, not embroidery."""
    px = 0
    # stitches = []
    while True:
        stitch_count = read_int_16le(f)
        block_length = read_int_16le(f)
        if block_length is None:
            return
        if block_length >= 256:
            break
        if stitch_count == 0:
            continue
        for i in range(0, stitch_count):
            x = read_int_8(f)
            y = read_int_8(f)
            if y > 16:
                y = -(32 - y)  # This is 5 bit signed number.
            if x > 32:
                x = -(64 - x)  # This is a 6 bit signed number.
            x *= 2.5
            y *= -2.5
            dx = x
            out.stitch_abs(px + x, y)  # This is a hybrid relative, absolute value.
            px += dx
    # f.seek(20,1)
    # length_table_index = read_int_8(f)
    # length_table_size = read_int_8(f)
    # length_table = []
    # for j in range(length_table_size):
    #     v1 = read_int_16le(f)
    #     v2 = read_int_16le(f)
    #     length_table.append((v1, v2))
    # print("Length Index is: %d -- %s" % (length_table_index, str(length_table[length_table_index])))
    # print(length_table)
    # width_table_index = read_int_8(f)
    # width_table_size = read_int_8(f)
    # width_table = []
    # for j in range(width_table_size):
    #     v1 = read_int_16le(f)
    #     v2 = read_int_16le(f)
    #     width_table.append((v1, v2))
    # print("Width Index is: %d -- %s" % (width_table_index, str(width_table[width_table_index])))
    # print(width_table)
    out.end()


def read(f, out, settings=None):
    f.seek(0x64, 0)
    read_pmv_stitches(f, out)



from .EmbThread import build_unique_palette
from .EmbConstant import *
from .EmbThreadPec import get_thread_set
from .PecGraphics import get_blank, draw_scaled
from .WriteHelper import write_int_8, write_int_16le, write_int_16be, \
    write_int_24le, write_string_utf8

SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_JUMP
FULL_JUMP = True
ROUND = True
MAX_JUMP_DISTANCE = 2047
MAX_STITCH_DISTANCE = 2047

MASK_07_BIT = 0b01111111
JUMP_CODE = 0b00010000
TRIM_CODE = 0b00100000
FLAG_LONG = 0b10000000
PEC_ICON_WIDTH = 48
PEC_ICON_HEIGHT = 38


def write(pattern, f, settings=None):
    pattern.fix_color_count()
    pattern.interpolate_stop_as_duplicate_color()
    f.write(bytes("#PEC0001".encode('utf8')))
    write_pec(pattern, f)


def write_pec(pattern, f, threadlist=None):
    extends = pattern.bounds()
    if threadlist is None:
        pattern.fix_color_count()
        threadlist = pattern.threadlist
    color_info = write_pec_header(pattern, f, threadlist)
    write_pec_block(pattern, f, extends)
    write_pec_graphics(pattern, f, extends)
    return color_info


def write_pec_header(pattern, f, threadlist):
    name = pattern.get_metadata("name", "Untitled")
    write_string_utf8(f, "LA:%-16s\r" % name[:8])
    f.write(b'\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\xFF\x00')
    write_int_8(f, int(PEC_ICON_WIDTH / 8))  # PEC BYTE STRIDE
    write_int_8(f, int(PEC_ICON_HEIGHT))  # PEC ICON HEIGHT

    thread_set = get_thread_set()

    color_index_list = build_unique_palette(thread_set, pattern.threadlist)

    rgb_list = [thread.color for thread in threadlist]

    current_thread_count = len(color_index_list)
    if current_thread_count != 0:
        f.write(b'\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20')
        add_value = current_thread_count - 1
        color_index_list.insert(0, add_value)
        f.write(bytes(bytearray(color_index_list)))
    else:
        f.write(b'\x20\x20\x20\x20\x64\x20\x00\x20\x00\x20\x20\x20\xFF')

    for i in range(current_thread_count, 463):
        f.write(b'\x20')  # 520
    return color_index_list, rgb_list


def write_pec_block(pattern, f, extends):
    width = extends[2] - extends[0]
    height = extends[3] - extends[1]

    stitch_block_start_position = f.tell()
    f.write(b'\x00\x00')
    write_int_24le(f, 0)  # Space holder.
    f.write(b'\x31\xff\xf0')
    write_int_16le(f, int(round(width)))
    write_int_16le(f, int(round(height)))
    write_int_16le(f, 0x1E0)
    write_int_16le(f, 0x1B0)

    write_int_16be(f, 0x9000 | -int(round(extends[0])))
    write_int_16be(f, 0x9000 | -int(round(extends[1])))

    pec_encode(pattern, f)

    stitch_block_length = f.tell() - stitch_block_start_position

    current_position = f.tell()
    f.seek(stitch_block_start_position + 2, 0)
    write_int_24le(f, stitch_block_length)
    f.seek(current_position, 0)


def write_pec_graphics(pattern, f, extends):
    blank = get_blank()
    for block in pattern.get_as_stitchblock():
        stitches = block[0]
        draw_scaled(extends, stitches, blank, 6, 4)
    f.write(bytes(bytearray(blank)))

    for block in pattern.get_as_colorblocks():
        stitches = [s for s in block[0] if s[2] == STITCH]
        blank = get_blank()  # [ 0 ] * 6 * 38
        draw_scaled(extends, stitches, blank, 6)
        f.write(bytes(bytearray(blank)))


def encode_long_form(value):
    value &= 0b0000111111111111
    value |= 0b1000000000000000
    return value


def flag_jump(long_form):
    return long_form | (JUMP_CODE << 8)


def flag_trim(long_form):
    return long_form | (TRIM_CODE << 8)


def pec_encode(pattern, f):
    color_two = True
    jumping = False
    stitches = pattern.stitches
    xx = 0
    yy = 0
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2] & COMMAND_MASK
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy
        if data == STITCH:
            if jumping and dx != 0 and dy != 0:
                f.write(b'\x00\x00')
                jumping = False
            if -64 < dx < 63 and -64 < dy < 63:
                f.write(bytes(bytearray([dx & MASK_07_BIT, dy & MASK_07_BIT])))
            else:
                dx = encode_long_form(dx)
                dy = encode_long_form(dy)
                data = [
                    (dx >> 8) & 0xFF,
                    dx & 0xFF,
                    (dy >> 8) & 0xFF,
                    dy & 0xFF]
                f.write(bytes(bytearray(data)))
            continue
        elif data == JUMP:
            jumping = True
            dx = encode_long_form(dx)
            dx = flag_trim(dx)
            dy = encode_long_form(dy)
            dy = flag_trim(dy)
            f.write(bytes(bytearray([
                (dx >> 8) & 0xFF,
                dx & 0xFF,
                (dy >> 8) & 0xFF,
                dy & 0xFF
            ])))
            continue
        elif data == COLOR_CHANGE:
            if jumping:
                f.write(b'\x00\x00')
                jumping = False
            f.write(b'\xfe\xb0')
            if color_two:
                f.write(b'\x02')
            else:
                f.write(b'\x01')
            color_two = not color_two
            continue
        elif data == STOP:
            continue  # These will already be processed into duplicate colors.
        elif data == TRIM:
            continue
        elif data == END:
            f.write(b'\xff')
            break



from .ReadHelper import signed8


def read_emd_stitches(f, out):
    count = 0
    while True:
        count += 1
        b = bytearray(f.read(2))
        if len(b) != 2:
            break

        if b[0] != 0x80:
            x = signed8(b[0])
            y = -signed8(b[1])
            out.stitch(x, y)
            continue
        control = b[1]
        if control == 0x80:
            b = bytearray(f.read(2))
            if len(b) != 2:
                break
            x = signed8(b[0])
            y = -signed8(b[1])
            out.move(x, y)
            continue
        if control == 0x2A:
            out.color_change()
            continue
        if control == 0x7D:
            continue  # Dunno, occurs at position 0.
        if control == 0xAD:
            out.trim()
            continue
        if control == 0x90:
            out.trim()  # Final command before returning to start.
            continue
        elif control == 0xFD:
            break
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    f.seek(0x30, 0)
    read_emd_stitches(f, out)



from .ExpReader import read_exp_stitches
from .ReadHelper import read_int_32le


def read(f, out, settings=None):
    # File starts with STX
    f.seek(0x0C, 1)
    color_start_position = read_int_32le(f)
    dunno_block_start_position = read_int_32le(f)
    stitch_start_position = read_int_32le(f)
    f.seek(stitch_start_position, 0)
    read_exp_stitches(f, out)



from .EmbCompress import expand
from .EmbThreadHus import get_thread_set
from .ReadHelper import signed8, signed16, read_int_32le, read_int_16le, read_string_8


def read(f, out, settings=None):
    magic_code = read_int_32le(f)
    number_of_stitches = read_int_32le(f)
    number_of_colors = read_int_32le(f)

    extend_pos_x = signed16(read_int_16le(f))
    extend_pos_y = signed16(read_int_16le(f))
    extend_neg_x = signed16(read_int_16le(f))
    extend_neg_y = signed16(read_int_16le(f))

    command_offset = read_int_32le(f)
    x_offset = read_int_32le(f)
    y_offset = read_int_32le(f)

    string_value = read_string_8(f, 8)

    unknown_16_bit = read_int_16le(f)

    hus_thread_set = get_thread_set()
    for i in range(0, number_of_colors):
        index = read_int_16le(f)
        out.add_thread(hus_thread_set[index])
    f.seek(command_offset, 0)
    command_compressed = bytearray(f.read(x_offset - command_offset))
    f.seek(x_offset, 0)
    x_compressed = bytearray(f.read(y_offset - x_offset))
    f.seek(y_offset, 0)
    y_compressed = bytearray(f.read())

    command_decompressed = expand(command_compressed, number_of_stitches)
    x_decompressed = expand(x_compressed, number_of_stitches)
    y_decompressed = expand(y_compressed, number_of_stitches)

    stitch_count = min(len(command_decompressed), len(x_decompressed), len(y_decompressed))

    for i in range(0, stitch_count):
        cmd = command_decompressed[i]
        x = signed8(x_decompressed[i])
        y = -signed8(y_decompressed[i])
        if cmd == 0x80:  # STITCH
            out.stitch(x, y)
        elif cmd == 0x81:  # JUMP
            out.move(x, y)
        elif cmd == 0x84:  # COLOR_CHANGE
            out.color_change(x, y)
        elif cmd == 0x88:  # TRIM
            if x != 0 or y != 0:
                out.move(x, y)
            out.trim()
        elif cmd == 0x90:  # END
            break
        else:  # UNMAPPED COMMAND
            break
    out.end()



from .ReadHelper import signed8


def read_exp_stitches(f, out):
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] != 0x80:
            x = signed8(b[0])
            y = -signed8(b[1])
            out.stitch(x, y)
            continue

        control = b[1]
        b = bytearray(f.read(2))  # 07 00
        if len(b) != 2:
            break
        x = signed8(b[0])
        y = -signed8(b[1])
        if control == 0x80:  # Trim
            out.trim()
            continue
        elif control == 0x02:
            out.stitch(x, y)
            # This shouldn't exist.
            continue
        elif control == 0x04:  # Jump
            out.move(x, y)
            continue
        elif control == 0x01:  # Colorchange
            out.color_change()
            if x != 0 or y != 0:
                out.move(x, y)
            continue
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    read_exp_stitches(f, out)



def signed8(b):
    if b > 127:
        return -256 + b
    else:
        return b


def signed16(v):
    v &= 0xFFFF
    if v > 0x7FFF:
        return - 0x10000 + v
    else:
        return v


def signed24(v):
    v &= 0xFFFFFF
    if v > 0x7FFFFF:
        return - 0x1000000 + v
    else:
        return v


def read_signed(stream, n):
    byte = bytearray(stream.read(n))
    signed_bytes = []
    for b in byte:
        signed_bytes.append(signed8(b))
    return signed_bytes


def read_sint_8(stream):
    byte = bytearray(stream.read(1))
    if len(byte) == 1:
        return signed8(byte[0])
    return None


def read_int_8(stream):
    byte = bytearray(stream.read(1))
    if len(byte) == 1:
        return byte[0]
    return None


def read_int_16le(stream):
    byte = bytearray(stream.read(2))
    if len(byte) == 2:
        return (byte[0] & 0xFF) + ((byte[1] & 0xFF) << 8)
    return None


def read_int_16be(stream):
    byte = bytearray(stream.read(2))
    if len(byte) == 2:
        return (byte[1] & 0xFF) + ((byte[0] & 0xFF) << 8)
    return None


def read_int_24le(stream):
    b = bytearray(stream.read(3))
    if len(b) == 3:
        return (b[0] & 0xFF) + ((b[1] & 0xFF) << 8) + \
               ((b[2] & 0xFF) << 16)
    return None


def read_int_24be(stream):
    b = bytearray(stream.read(3))
    if len(b) == 3:
        return (b[2] & 0xFF) + ((b[1] & 0xFF) << 8) + \
               ((b[0] & 0xFF) << 16)
    return None


def read_int_32le(stream):
    b = bytearray(stream.read(4))
    if len(b) == 4:
        return (b[0] & 0xFF) + ((b[1] & 0xFF) << 8) + \
               ((b[2] & 0xFF) << 16) + ((b[3] & 0xFF) << 24)
    return None


def read_int_32be(stream):
    b = bytearray(stream.read(4))
    if len(b) == 4:
        return (b[3] & 0xFF) + ((b[2] & 0xFF) << 8) + \
               ((b[1] & 0xFF) << 16) + ((b[0] & 0xFF) << 24)
    return None


def read_string_8(stream, length):
    byte = stream.read(length)
    try:
        return byte.decode('utf8')
    except UnicodeDecodeError:
        return None  # Must be > 128 chars.


def read_string_16(stream, length):
    byte = stream.read(length)
    try:
        return byte.decode('utf16')
    except UnicodeDecodeError:
        return None



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyembroidery",
    version="1.4.26",
    author="Tatarize",
    author_email="tatarize@gmail.com",
    description="Embroidery IO library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EmbroidePy/pyembroidery",
    packages=setuptools.find_packages(),
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Software Development :: Libraries :: Python Modules'
    ),
)


# The commands below 0xFF are intended to denote proper commands.
# The encodings beyond this is supplying additional information.

COMMAND_MASK = 0x000000FF
THREAD_MASK = 0x0000FF00
NEEDLE_MASK = 0x00FF0000
ORDER_MASK = 0xFF000000

FLAGS_MASK = 0x0000FF00

NO_COMMAND = -1
STITCH = 0
JUMP = 1
TRIM = 2
STOP = 3
END = 4
COLOR_CHANGE = 5
NEEDLE_SET = 9
SEQUIN_MODE = 6
SEQUIN_EJECT = 7
SLOW = 0xB  # 11
FAST = 0xC  # 12

# Assigns the change sequences, allowing the color change commands to trigger any event.
# For these can be processed in some needle, thread, order values to preset or postset the
# Underlying change sequence.
SET_CHANGE_SEQUENCE = 0x10  # 16

# Stitch with implied contingency.
SEW_TO = 0xB0  # 176
NEEDLE_AT = 0xB1  # 177

STITCH_BREAK = 0xE0  # 224

SEQUENCE_BREAK = 0xE1  # 225
COLOR_BREAK = 0xE2  # 226

# Middle Level commands.
TIE_ON = 0xE4  # 228
TIE_OFF = 0xE5  # 229
FRAME_EJECT = 0xE9  # 233

# Matrix Commands.
MATRIX_TRANSLATE = 0xC0  # 192
MATRIX_SCALE_ORIGIN = 0xC1  # 193
MATRIX_ROTATE_ORIGIN = 0xC2  # 194
MATRIX_SCALE = 0xC4  # 196
MATRIX_ROTATE = 0xC5  # 197
MATRIX_RESET = 0xC3  # 195

OPTION_MAX_STITCH_LENGTH = 0xD5
OPTION_MAX_JUMP_LENGTH = 0xD6
OPTION_EXPLICIT_TRIM = 0xD7
OPTION_IMPLICIT_TRIM = 0xD8

CONTINGENCY_TIE_ON_NONE = 0xD3
CONTINGENCY_TIE_ON_THREE_SMALL = 0xD1

CONTINGENCY_TIE_OFF_NONE = 0xD4
CONTINGENCY_TIE_OFF_THREE_SMALL = 0xD2

CONTINGENCY_LONG_STITCH_NONE = 0xF0
CONTINGENCY_LONG_STITCH_JUMP_NEEDLE = 0xF1
CONTINGENCY_LONG_STITCH_SEW_TO = 0xF2

CONTINGENCY_SEQUIN_UTILIZE = 0xF5
CONTINGENCY_SEQUIN_JUMP = 0xF6
CONTINGENCY_SEQUIN_STITCH = 0xF7
CONTINGENCY_SEQUIN_REMOVE = 0xF8

ALTERNATIVE = 0x100  # Generic flag for an alternative form.



from .EmbThread import EmbThread


def get_thread_set():
    return [
        EmbThreadHus("#000000", "Black", "026"),
        EmbThreadHus("#0000e7", "Blue", "005"),
        EmbThreadHus("#00c600", "Green", "002"),
        EmbThreadHus("#ff0000", "Red", "014"),
        EmbThreadHus("#840084", "Purple", "008"),
        EmbThreadHus("#ffff00", "Yellow", "020"),
        EmbThreadHus("#848484", "Grey", "024"),
        EmbThreadHus("#8484e7", "Light Blue", "006"),
        EmbThreadHus("#00ff84", "Light Green", "003"),
        EmbThreadHus("#ff7b31", "Orange", "017"),
        EmbThreadHus("#ff8ca5", "Pink", "011"),
        EmbThreadHus("#845200", "Brown", "028"),
        EmbThreadHus("#ffffff", "White", "022"),
        EmbThreadHus("#000084", "Dark Blue", "004"),
        EmbThreadHus("#008400", "Dark Green", "001"),
        EmbThreadHus("#7b0000", "Dark Red", "013"),
        EmbThreadHus("#ff6384", "Light Red", "015"),
        EmbThreadHus("#522952", "Dark Purple", "007"),
        EmbThreadHus("#ff00ff", "Light Purple", "009"),
        EmbThreadHus("#ffde00", "Dark Yellow", "019"),
        EmbThreadHus("#ffff9c", "Light Yellow", "021"),
        EmbThreadHus("#525252", "Dark Grey", "025"),
        EmbThreadHus("#d6d6d6", "Light Grey", "023"),
        EmbThreadHus("#ff5208", "Dark Orange", "016"),
        EmbThreadHus("#ff9c5a", "Light Orange", "018"),
        EmbThreadHus("#ff52b5", "Dark Pink", "010"),
        EmbThreadHus("#ffc6de", "Light Pink", "012"),
        EmbThreadHus("#523100", "Dark Brown", "027"),
        EmbThreadHus("#b5a584", "Light Brown", "029")
    ]


class EmbThreadHus(EmbThread):
    def __init__(self, color, description, catalog_number=None):
        EmbThread.__init__(self)
        self.set(color)
        self.description = description
        self.catalog_number = catalog_number
        self.brand = "Hus"
        self.chart = "Hus"




from __future__ import print_function

import unittest

from pyembroidery import *
from pyembroidery.EmbThreadPec import *
from pyembroidery.EmbThread import build_unique_palette, build_nonrepeat_palette, build_palette


class TestPalettes(unittest.TestCase):

    def test_unique_palette(self):
        """Similar elements should not plot to the same palette index"""
        pattern = EmbPattern()
        pattern += "#FF0001"
        pattern += "Blue"
        pattern += "Blue"
        pattern += "Red"
        threadset = get_thread_set()
        palette = build_unique_palette(threadset,pattern.threadlist)
        self.assertNotEqual(palette[0],palette[3])  # Red and altered Red
        self.assertEqual(palette[1], palette[2])  # Blue and Blue

    def test_unique_palette_large(self):
        """Excessive palette entries that all map, should be mapped"""
        pattern = EmbPattern()
        for x in range(0, 100):
            pattern += "black"
        threadset = get_thread_set()
        palette = build_unique_palette(threadset, pattern.threadlist)
        self.assertEqual(palette[0], palette[1])

    def test_unique_palette_unmap(self):
        """Excessive palette entries can't all map, should map what it can and repeat"""
        pattern = EmbPattern()
        for i in range(0, 100):
            thread = EmbThread()
            thread.set_color(i, i, i)
            pattern += thread
        threadset = get_thread_set()
        palette = build_unique_palette(threadset, pattern.threadlist)
        palette.sort()

    def test_unique_palette_max(self):
        """If the entries equal the list they should all map."""
        pattern = EmbPattern()
        threadset = get_thread_set()
        for i in range(0, len(threadset)-2):
            thread = EmbThread()
            thread.set_color(i, i, i)
            pattern += thread
        palette = build_unique_palette(threadset, pattern.threadlist)
        palette.sort()
        for i in range(1,len(palette)):
            self.assertNotEqual(palette[i-1], palette[i])

    def test_nonrepeat_palette_moving(self):
        """The almost same color should not get plotted to the same palette index"""
        pattern = EmbPattern()
        pattern += "Red"
        pattern += "Blue"
        pattern += "#0100FF"
        pattern += "Red"
        threadset = get_thread_set()
        palette = build_nonrepeat_palette(threadset,pattern.threadlist)
        self.assertEqual(palette[0],palette[3]) # Red and Red
        self.assertNotEqual(palette[1], palette[2])  # Blue and altered Blue

    def test_nonrepeat_palette_stay_moved(self):
        """An almost same moved, only temporary"""
        pattern = EmbPattern()
        pattern += "Red"
        pattern += "Blue"
        pattern += "#0100FF"
        pattern += "Red"
        pattern += "#0100FF"
        threadset = get_thread_set()
        palette = build_nonrepeat_palette(threadset,pattern.threadlist)
        self.assertEqual(palette[0],palette[3]) # Red and Red
        self.assertNotEqual(palette[1], palette[2])  # Blue and altered Blue
        self.assertNotEqual(palette[2], palette[4])  # same color, but color was moved

    def test_nonrepeat_palette_same(self):
        """The same exact same color if repeated should remain"""
        pattern = EmbPattern()
        pattern += "Red"
        pattern += "Blue"
        pattern += "#0000FF"  # actual blue
        pattern += "Red"
        threadset = get_thread_set()
        palette = build_nonrepeat_palette(threadset,pattern.threadlist)
        self.assertEqual(palette[0],palette[3]) # Red and Red
        self.assertEqual(palette[1], palette[2])  # Blue and Blue

    def test_palette(self):
        """Similar colors map to the same index"""
        pattern = EmbPattern()
        pattern += "#FF0001"
        pattern += "Blue"
        pattern += "Blue"
        pattern += "Red"
        threadset = get_thread_set()
        palette = build_palette(threadset,pattern.threadlist)
        self.assertEqual(palette[0],palette[3])  # Red and altered Red
        self.assertEqual(palette[1], palette[2])  # Blue and Blue



from .EmbFunctions import *
from .WriteHelper import write_int_16le, write_int_32le

THREAD_CHANGE_COMMAND = NEEDLE_SET
SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_JUMP
WRITES_SPEEDS = True
FULL_JUMP = False
MAX_JUMP_DISTANCE = 127
MAX_STITCH_DISTANCE = 127


def write(pattern, f, settings=None):
    stitches = pattern.stitches
    stitch_count = len(stitches)
    for i in range(0, 0x80):
        f.write(b'0')
    if stitch_count == 0:
        return
    extends = pattern.bounds()
    write_int_16le(f, int(extends[0]))
    write_int_16le(f, -int(extends[3]))
    write_int_16le(f, int(extends[2]))
    write_int_16le(f, -int(extends[1]))
    write_int_32le(f, 0)  # Dunno.

    write_int_32le(f, stitch_count + 1)
    last_stitch = stitches[stitch_count - 1]
    write_int_16le(f, int(last_stitch[0]))
    write_int_16le(f, -int(last_stitch[1]))
    for i in range(f.tell(), 0x100):
        f.write(b'\x00')
    xx = 0
    yy = 0
    trigger_fast = False
    trigger_slow = False
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2] & COMMAND_MASK
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy
        if data == SLOW:
            trigger_slow = True
            continue
        if data == FAST:
            trigger_fast = True
            continue
        cmd = 0x80
        if dy >= 0:
            cmd |= 0x40
        if dx <= 0:
            cmd |= 0x20
        delta_x = abs(dx)
        delta_y = abs(dy)
        if data == STITCH:
            if trigger_fast:
                trigger_fast = False
                cmd |= 0x02
            if trigger_slow:
                trigger_slow = False
                cmd |= 0x04
            f.write(bytes(bytearray([cmd, delta_y, delta_x])))
        elif data == JUMP:  # If you did both FAST, SLOW, and JUMP, you'd get a trim.
            if trigger_fast:
                trigger_fast = False
                cmd |= 0x02
            if trigger_slow:
                trigger_slow = False
                cmd |= 0x04
            cmd |= 0x01
            f.write(bytes(bytearray([cmd, delta_y, delta_x])))
        elif data == STOP:
            cmd |= 0x08
            f.write(bytes(bytearray([cmd, delta_y, delta_x])))
        elif data == TRIM:
            cmd |= 0x07
            f.write(bytes(bytearray([cmd, delta_y, delta_x])))
        elif data == NEEDLE_SET:
            decoded = decode_embroidery_command(stitch[2])
            needle = decoded[2]
            if needle >= 15:
                needle = (needle % 15) + 1
            cmd |= 0x08
            cmd += needle
            f.write(bytes(bytearray([cmd, delta_y, delta_x])))
        elif data == END:
            break
    f.write(b'\xF8\x00\x00')



from .EmbFunctions import *
from .WriteHelper import write_string_utf8

SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_STITCH
SCALE = (-1, -1)  # This performs a default X,Y flip.


def write_data(pattern, f):
    bounds = [float(e) / 10.0 for e in pattern.bounds()]  # convert to mm.
    width = bounds[2] - bounds[0]
    height = bounds[3] - bounds[1]
    count_stitches = pattern.count_stitches()
    count_threads = pattern.count_color_changes()

    write_string_utf8(f, '(STITCH_COUNT: %d)\n' % count_stitches)
    write_string_utf8(f, '(THREAD_COUNT: %d)\n' % count_threads)
    write_string_utf8(f, '(EXTENTS_LEFT: %.3f)\n' % bounds[0])
    write_string_utf8(f, '(EXTENTS_TOP: %.3f)\n' % bounds[1])
    write_string_utf8(f, '(EXTENTS_RIGHT: %.3f)\n' % bounds[2])
    write_string_utf8(f, '(EXTENTS_BOTTOM: %.3f)\n' % bounds[3])
    write_string_utf8(f, '(EXTENTS_WIDTH: %.3f)\n' % width)
    write_string_utf8(f, '(EXTENTS_HEIGHT: %.3f)\n' % height)

    stitch_counts = {}
    for s in pattern.stitches:
        command = s[2] & COMMAND_MASK
        if command in stitch_counts:
            stitch_counts[command] += 1
        else:
            stitch_counts[command] = 1

    names = get_common_name_dictionary()
    if len(stitch_counts) != 0:
        for the_key, the_value in stitch_counts.items():
            try:
                the_key &= COMMAND_MASK
                name = "COMMAND_" + names[the_key]
            except (IndexError, KeyError):
                name = "COMMAND_UNKNOWN_" + str(the_key)
            write_string_utf8(f, '(%s: %d)\n' % (name, the_value))


def write_metadata(pattern, f):
    if len(pattern.extras) > 0:
        for the_key, the_value in pattern.extras.items():
            try:
                if isinstance(the_value, basestring):
                    write_string_utf8(f, '(%s: %s)\n' % (the_key, the_value))
            except NameError:
                if isinstance(the_value, str):
                    write_string_utf8(f, '(%s: %s)\n' % (the_key, the_value))


def write_threads(pattern, f):
    if len(pattern.threadlist) > 0:
        for i, thread in enumerate(pattern.threadlist):
            write_string_utf8(f, '(Thread%d: %s %s %s %s)\n' % (
                i,
                thread.hex_color(),
                thread.description,
                thread.brand,
                thread.catalog_number,
            ))


def write(pattern, f, settings=None):
    if settings is None:
        settings = {}
    increment_value = settings.get('stitch_z_travel', 10.0)
    write_data(pattern, f)
    write_metadata(pattern, f)
    write_threads(pattern, f)

    z = 0.0
    for i, stitch in enumerate(pattern.stitches):
        x = float(stitch[0]) / 10.0
        y = float(stitch[1]) / 10.0
        cmd = decode_embroidery_command(stitch[2])
        command = cmd[0]
        if command == STITCH:
            write_string_utf8(f, 'G00 X%.3f Y%.3f\nG00 Z%.1f\n' % (x, y, z))
            z += increment_value
            continue
        elif command == JUMP:
            continue
        elif command == TRIM:
            continue
        elif command == COLOR_CHANGE:
            write_string_utf8(f, 'M00\n')
            continue
        elif command == STOP:
            write_string_utf8(f, 'M00\n')
            continue
        elif command == END:
            write_string_utf8(f, 'M30\n')
            continue
        break  # Code shouldn't reach this point.



import math


class EmbMatrix:
    def __init__(self, m=None):
        if m is None:
            self.m = self.get_identity()
        else:
            self.m = m

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        return self.m == other.m

    def __matmul__(self, other):
        return EmbMatrix(EmbMatrix.matrix_multiply(self.m, other.m))

    def __rmatmul__(self, other):
        return EmbMatrix(EmbMatrix.matrix_multiply(self.m, other.m))

    def __imatmul__(self, other):
        self.m = EmbMatrix.matrix_multiply(self.m, other.m)

    def __str__(self):
        return "[%3f, %3f, %3f\n %3f, %3f, %3f\n %3f, %3f, %3f]" % self.m

    def get_matrix(self):
        return self.m

    def reset(self):
        self.m = self.get_identity()

    def inverse(self):
        m = self.m
        m48s75 = m[4] * m[8] - m[7] * m[5]
        m38s56 = m[5] * m[6] - m[3] * m[8]
        m37s46 = m[3] * m[7] - m[4] * m[6]
        det = m[0] * m48s75 + m[1] * m38s56 + m[2] * m37s46
        inverse_det = 1.0 / float(det)
        self.m = [
            m48s75 * inverse_det,
            (m[2] * m[7] - m[1] * m[8]) * inverse_det,
            (m[1] * m[5] - m[2] * m[4]) * inverse_det,
            m38s56 * inverse_det,
            (m[0] * m[8] - m[2] * m[6]) * inverse_det,
            (m[3] * m[2] - m[0] * m[5]) * inverse_det,
            m37s46 * inverse_det,
            (m[6] * m[1] - m[0] * m[7]) * inverse_det,
            (m[0] * m[4] - m[3] * m[1]) * inverse_det,
        ]

    def post_scale(self, sx=1, sy=None, x=0, y=0):
        if sy is None:
            sy = sx
        if x is None:
            x = 0
        if y is None:
            y = 0
        if x == 0 and y == 0:
            self.m = self.matrix_multiply(self.m, self.get_scale(sx, sy))
        else:
            self.post_translate(x, y)
            self.post_scale(sx, sy)
            self.post_translate(-x, -y)

    def post_translate(self, tx, ty):
        self.m = self.matrix_multiply(self.m, self.get_translate(tx, ty))

    def post_rotate(self, theta, x=0, y=0):
        if x is None:
            x = 0
        if y is None:
            y = 0
        if x == 0 and y == 0:
            self.m = self.matrix_multiply(self.m, self.get_rotate(theta))
        else:
            self.post_translate(x, y)
            self.post_rotate(theta)
            self.post_translate(-x, -y)

    def post_cat(self, matrix_list):
        for mx in matrix_list:
            self.m = self.matrix_multiply(self.m, mx)

    def pre_scale(self, sx=1, sy=None):
        if sy is None:
            sy = sx
        self.m = self.matrix_multiply(self.get_scale(sx, sy), self.m)

    def pre_translate(self, tx, ty):
        self.m = self.matrix_multiply(self.get_translate(tx, ty), self.m)

    def pre_rotate(self, theta):
        self.m = self.matrix_multiply(self.get_rotate(theta), self.m)

    def pre_cat(self, matrix_list):
        for mx in matrix_list:
            self.m = self.matrix_multiply(mx, self.m)

    def point_in_matrix_space(self, v0, v1=None):
        m = self.m
        if v1 is None:
            try:
                return [
                    v0[0] * m[0] + v0[1] * m[3] + 1 * m[6],
                    v0[0] * m[1] + v0[1] * m[4] + 1 * m[7],
                    v0[2]
                ]
            except IndexError:
                return [
                    v0[0] * m[0] + v0[1] * m[3] + 1 * m[6],
                    v0[0] * m[1] + v0[1] * m[4] + 1 * m[7]
                    # Must not have had a 3rd element.
                ]
        return [
            v0 * m[0] + v1 * m[3] + 1 * m[6],
            v0 * m[1] + v1 * m[4] + 1 * m[7]
        ]

    def apply(self, v):
        m = self.m
        nx = v[0] * m[0] + v[1] * m[3] + 1 * m[6]
        ny = v[0] * m[1] + v[1] * m[4] + 1 * m[7]
        v[0] = nx
        v[1] = ny

    @staticmethod
    def get_identity():
        return \
            1, 0, 0, \
            0, 1, 0, \
            0, 0, 1  # identity

    @staticmethod
    def get_scale(sx, sy=None):
        if sy is None:
            sy = sx
        return \
            sx, 0, 0, \
            0, sy, 0, \
            0, 0, 1

    @staticmethod
    def get_translate(tx, ty):
        return \
            1, 0, 0, \
            0, 1, 0, \
            tx, ty, 1

    @staticmethod
    def get_rotate(theta):
        tau = math.pi * 2
        theta *= tau / 360
        ct = math.cos(theta)
        st = math.sin(theta)
        return \
            ct, st, 0, \
            -st, ct, 0, \
            0, 0, 1

    @staticmethod
    def matrix_multiply(m0, m1):
        return [
            m1[0] * m0[0] + m1[1] * m0[3] + m1[2] * m0[6],
            m1[0] * m0[1] + m1[1] * m0[4] + m1[2] * m0[7],
            m1[0] * m0[2] + m1[1] * m0[5] + m1[2] * m0[8],
            m1[3] * m0[0] + m1[4] * m0[3] + m1[5] * m0[6],
            m1[3] * m0[1] + m1[4] * m0[4] + m1[5] * m0[7],
            m1[3] * m0[2] + m1[4] * m0[5] + m1[5] * m0[8],
            m1[6] * m0[0] + m1[7] * m0[3] + m1[8] * m0[6],
            m1[6] * m0[1] + m1[7] * m0[4] + m1[8] * m0[7],
            m1[6] * m0[2] + m1[7] * m0[5] + m1[8] * m0[8]]
